from __future__ import absolute_import
import unittest
import tarfile
import os
import json

from yhat.deployment.save_session import save_function
from yhat import batch

# Function test whether the bundling code will
def external_func():
    print("Hello from outside the class!")

class BatchTestCase(unittest.TestCase):

    test_archive = ".tmp_yhat_job_test.tar.gz"

    def setUp(self):
        # Create local file specifying mock install instructions
        print("creating tar...")
        with open("yhat.yaml", "w") as f:
            f.write("yum install lsof\necho 'Installs complete!'")
        # Create requirements.txt with mock python dependencies
        with open("requirements.txt", "w") as f:
            f.write("nose==1.3.7\nmock==1.3.0")

    # Remove test files that were created locally
    def tearDown(self):
        os.remove(self.test_archive)

        files = ["yhat.yaml", "requirements.txt"]
        for f in files:
            if os.path.isfile(f):
                os.remove(f)

    # Create a new batch job class with an execute method
    class TestBatchJob(batch.BatchJob):
        def execute():
            print("Hello")
            external_func()

    # Test for creation of the bundle tar
    def test_create_bundle_tar(self):
        batch_job = self.TestBatchJob("test_job", username="bob", \
            apikey="123", url="http://localhost:9000")
        bundle = save_function(batch_job.__class__, globals())
        bundle_str = json.dumps(bundle)

        print("checking tar contents...")
        # Create the bundle tar using the function from batch
        batch_job._BatchJob__create_bundle_tar(bundle_str, \
            self.test_archive)

        # Make sure the tar file includes all the correct files
        # and contents
        with tarfile.open(self.test_archive, "r:gz") as f:
            requirements = f.extractfile("requirements.txt").read()
            print("checking requirements...")
            self.assertIn(b"nose==1.3.7", requirements)
            self.assertIn(b"mock==1.3.0", requirements)

            print("checking yaml...")
            yaml_file = f.extractfile("yhat.yaml").read()
            self.assertIn(b"Installs complete!", yaml_file)
            self.assertIn(b"yum install lsof", yaml_file)

            print("checking bundle...")
            bundle = f.extractfile("bundle.json").read()
            self.assertIn(b"""print(\\"Hello\\")""", bundle)
            self.assertIn(b"Hello from outside the class!", bundle)

if __name__ == '__main__':
    unittest.main()



import os

def detect_explicit_submodules(model_object):
    submodules = []
    files = getattr(model_object, "FILES", [])
    for f in files:
        basename = os.path.basename(f)
        parent_dir = os.path.dirname(f)
        source = open(f, 'rb').read()
        submodule = {
            "parent_dir": parent_dir,
            "name": basename,
            "source": source
        }
        submodules.append(submodule)

        directories = parent_dir.split('/')
        for i in range(len(directories)):
            submodules.append({
                "parent_dir": "/".join(directories[:i]),
                "name": "__init__.py",
                "source": ""
            })

    return submodules



from __future__ import print_function
import tokenize
import os
import shutil
import sys

verbose = False
recurse = False
dryrun = False
makebackup = True


def usage(msg=None):
    if msg is None:
        msg = __doc__
    print(msg, file=sys.stderr)


def errprint(*args):
    sys.stderr.write(" ".join(str(arg) for arg in args))
    sys.stderr.write("\n")


def main():
    import getopt
    global verbose, recurse, dryrun, makebackup
    try:
        opts, args = getopt.getopt(sys.argv[1:], "drnvh",
                        ["dryrun", "recurse", "nobackup", "verbose", "help"])
    except getopt.error as msg:
        usage(msg)
        return
    for o, a in opts:
        if o in ('-d', '--dryrun'):
            dryrun = True
        elif o in ('-r', '--recurse'):
            recurse = True
        elif o in ('-n', '--nobackup'):
            makebackup = False
        elif o in ('-v', '--verbose'):
            verbose = True
        elif o in ('-h', '--help'):
            usage()
            return
    if not args:
        r = Reindenter(sys.stdin)
        r.run()
        r.write(sys.stdout)
        return
    for arg in args:
        check(arg)


def check(file):
    if os.path.isdir(file) and not os.path.islink(file):
        if verbose:
            print("listing directory", file)
        names = os.listdir(file)
        for name in names:
            fullname = os.path.join(file, name)
            if ((recurse and os.path.isdir(fullname) and
                 not os.path.islink(fullname) and
                 not os.path.split(fullname)[1].startswith("."))
                or name.lower().endswith(".py")):
                check(fullname)
        return

    if verbose:
        print("checking", file, "...", end=' ')
    with open(file, 'rb') as f:
        encoding, _ = tokenize.detect_encoding(f.readline)
    try:
        with open(file, encoding=encoding) as f:
            r = Reindenter(f)
    except IOError as msg:
        errprint("%s: I/O Error: %s" % (file, str(msg)))
        return

    newline = r.newlines
    if isinstance(newline, tuple):
        errprint("%s: mixed newlines detected; cannot process file" % file)
        return

    if r.run():
        if verbose:
            print("changed.")
            if dryrun:
                print("But this is a dry run, so leaving it alone.")
        if not dryrun:
            bak = file + ".bak"
            if makebackup:
                shutil.copyfile(file, bak)
                if verbose:
                    print("backed up", file, "to", bak)
            with open(file, "w", encoding=encoding, newline=newline) as f:
                r.write(f)
            if verbose:
                print("wrote new", file)
        return True
    else:
        if verbose:
            print("unchanged.")
        return False


def _rstrip(line, JUNK='\n \t'):
    """Return line stripped of trailing spaces, tabs, newlines.
    Note that line.rstrip() instead also strips sundry control characters,
    but at least one known Emacs user expects to keep junk like that, not
    mentioning Barry by name or anything <wink>.
    """

    i = len(line)
    while i > 0 and line[i - 1] in JUNK:
        i -= 1
    return line[:i]


class Reindenter:

    def __init__(self, f):
        self.find_stmt = 1  # next token begins a fresh stmt?
        self.level = 0      # current indent level

        # Raw file lines.
        self.raw = f.readlines()

        # File lines, rstripped & tab-expanded.  Dummy at start is so
        # that we can use tokenize's 1-based line numbering easily.
        # Note that a line is all-blank iff it's "\n".
        self.lines = [_rstrip(line).expandtabs() + "\n"
                      for line in self.raw]
        self.lines.insert(0, None)
        self.index = 1  # index into self.lines of next line

        # List of (lineno, indentlevel) pairs, one for each stmt and
        # comment line.  indentlevel is -1 for comment lines, as a
        # signal that tokenize doesn't know what to do about them;
        # indeed, they're our headache!
        self.stats = []

        # Save the newlines found in the file so they can be used to
        #  create output without mutating the newlines.
        self.newlines = f.newlines

    def run(self):
        tokens = tokenize.generate_tokens(self.getline)
        for _token in tokens:
            self.tokeneater(*_token)
        # Remove trailing empty lines.
        lines = self.lines
        while lines and lines[-1] == "\n":
            lines.pop()
        # Sentinel.
        stats = self.stats
        stats.append((len(lines), 0))
        # Map count of leading spaces to # we want.
        have2want = {}
        # Program after transformation.
        after = self.after = []
        # Copy over initial empty lines -- there's nothing to do until
        # we see a line with *something* on it.
        i = stats[0][0]
        after.extend(lines[1:i])
        for i in range(len(stats) - 1):
            thisstmt, thislevel = stats[i]
            nextstmt = stats[i + 1][0]
            have = getlspace(lines[thisstmt])
            want = thislevel * 4
            if want < 0:
                # A comment line.
                if have:
                    # An indented comment line.  If we saw the same
                    # indentation before, reuse what it most recently
                    # mapped to.
                    want = have2want.get(have, -1)
                    if want < 0:
                        # Then it probably belongs to the next real stmt.
                        for j in range(i + 1, len(stats) - 1):
                            jline, jlevel = stats[j]
                            if jlevel >= 0:
                                if have == getlspace(lines[jline]):
                                    want = jlevel * 4
                                break
                    if want < 0:           # Maybe it's a hanging
                                           # comment like this one,
                        # in which case we should shift it like its base
                        # line got shifted.
                        for j in range(i - 1, -1, -1):
                            jline, jlevel = stats[j]
                            if jlevel >= 0:
                                want = have + (getlspace(after[jline - 1]) -
                                               getlspace(lines[jline]))
                                break
                    if want < 0:
                        # Still no luck -- leave it alone.
                        want = have
                else:
                    want = 0
            assert want >= 0
            have2want[have] = want
            diff = want - have
            if diff == 0 or have == 0:
                after.extend(lines[thisstmt:nextstmt])
            else:
                for line in lines[thisstmt:nextstmt]:
                    if diff > 0:
                        if line == "\n":
                            after.append(line)
                        else:
                            after.append(" " * diff + line)
                    else:
                        remove = min(getlspace(line), -diff)
                        after.append(line[remove:])
        return self.raw != self.after

    def write(self, f):
        f.writelines(self.after)

    # Line-getter for tokenize.
    def getline(self):
        if self.index >= len(self.lines):
            line = ""
        else:
            line = self.lines[self.index]
            self.index += 1
        return line

    # Line-eater for tokenize.
    def tokeneater(self, type, token, slinecol, end, line,
                   INDENT=tokenize.INDENT,
                   DEDENT=tokenize.DEDENT,
                   NEWLINE=tokenize.NEWLINE,
                   COMMENT=tokenize.COMMENT,
                   NL=tokenize.NL):

        if type == NEWLINE:
            # A program statement, or ENDMARKER, will eventually follow,
            # after some (possibly empty) run of tokens of the form
            #     (NL | COMMENT)* (INDENT | DEDENT+)?
            self.find_stmt = 1

        elif type == INDENT:
            self.find_stmt = 1
            self.level += 1

        elif type == DEDENT:
            self.find_stmt = 1
            self.level -= 1

        elif type == COMMENT:
            if self.find_stmt:
                self.stats.append((slinecol[0], -1))
                # but we're still looking for a new stmt, so leave
                # find_stmt alone

        elif type == NL:
            pass

        elif self.find_stmt:
            # This is the first "real token" following a NEWLINE, so it
            # must be the first token of the next program statement, or an
            # ENDMARKER.
            self.find_stmt = 0
            if line:   # not endmarker
                self.stats.append((slinecol[0], self.level))


# Count number of leading blanks.
def getlspace(line):
    i, n = 0, len(line)
    while i < n and line[i] == " ":
        i += 1
    return i


if __name__ == '__main__':
    main()



from __future__ import print_function

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

from builtins import input
from builtins import bytes
import sys
import warnings
import base64
import json
import pickle
import terragon
import inspect
import tempfile
import zlib
import logging
import re
import os
import os.path
import requests
from six import string_types
from requests_toolbelt.multipart.encoder import MultipartEncoderMonitor, MultipartEncoder
from progressbar import ProgressBar, Percentage, Bar, FileTransferSpeed, ETA

from .submodules import detect_explicit_submodules
from .deployment.models import YhatModel
from .deployment.save_session import save_function, _get_source, reindent
from .utils import sizeof_fmt, is_valid_json

devnull = open(os.devnull, "w")

# If we can't import everything we need to detect requirements from
# this version of pip, then we just warn and turn off the feature.
getExplicitRequirements, getImplicitRequirements = None, None
try:
    from .requirements import getExplicitRequirements, getImplicitRequirements
except ImportError:
    warnings.warn("Unable to use this version of pip. Requirements detection disabled. Consider upgrading pip.")
    DETECT_REQUIREMENTS = False
else:
    DETECT_REQUIREMENTS = True



BASE_URI = "http://api.yhathq.com/"


class API(object):

    """
    An abstract class that implements some of the more generic methods for
    interacting with REST APIs.

    Parameters
    ----------
    base_uri: string
        URI of the API that you're working with
    """

    def __init__(self, base_uri):
        self.base_uri = base_uri
        self.headers = {'Content-Type': 'application/json'}

    def _get(self, endpoint, params):
        """
        Parameters
        ----------
        endpoint: string
            name of the rest endpoint you want to hit
        params: dictionary
            querystring parameters for API call

        Returns
        -------
        data: dictionary
            whatever is returned by the API
        """
        try:
            url = self.base_uri + endpoint + "?" + urlencode(params)
            auth = '%s:%s' % (params['username'], params['apikey'])
            base64string = base64.encodestring(auth).replace('\n', '')
            base64header = "Basic %s" % base64string
            headers = {
                'Content-Type': 'application/json',
                'Authorization': base64header
            }
            response = requests.get(url=url, headers=headers)
            rsp = response.text
            return json.loads(rsp)
        except Exception as e:
            raise e

    def _post(self, endpoint, params, data):
        """
        Parameters
        ----------
        endpoint: string
            name of the rest endpoint you want to hit
        params: dictionary
            querystring parameters for API call
        data: dictionary
            data you want transfered as JSON

        Returns
        -------
        data: dictionary
            whatever is returned by the API
        """
        if is_valid_json(data)==False:
            raise Exception("`data` is not valid JSON")

        url = self.base_uri + endpoint
        headers = { "Content-Type": "application/json"}
        username, apikey = params['username'], params['apikey']

        response = requests.post(url=url, headers=headers, data=data, params=params, auth=(username, apikey))
        if response.status_code > 200:
            # if we don't get a 200, write the error and raise an status Exception
            sys.stderr.write("error: " + response.text)
            response.raise_for_status()
        return response.json()

    def _post_file(self, endpoint, params, data, pb=True):

        # headers contains the necessary Content-Type and Content-Length
        # datagen is a generator object that yields the encoded parameters
        f = tempfile.NamedTemporaryFile(mode='wb', prefix='tmp_yhat_', delete=False)
        model_name = data['modelname'] + ".yhat"
        try:
            data = json.dumps(data)
        except UnicodeDecodeError as e:
            raise Exception("Could not serialize into JSON. String is not utf-8 encoded `%s`" % str(e.args[1]))
        zlib_compress(data, f)
        f.close()

        def createCallback(encoder):
            # Stuff for progress bar setup
            widgets = ['Transfering Model: ', Bar(), Percentage(), ' ', ETA(), ' ', FileTransferSpeed()]
            pbar = ProgressBar(max_value=encoder.len, widgets=widgets).start()
            def callback(monitor):
                current = monitor.bytes_read
                pbar.update(current)
            return callback

        encoder = MultipartEncoder(
            fields={'model_name': (open(f.name, 'rb'))}
        )

        username, apikey = params['username'], params['apikey']
        headers = {'Content-Type': encoder.content_type,}

        callback = createCallback(encoder)
        monitor = MultipartEncoderMonitor(encoder, callback)
        url = self.base_uri + endpoint

        # Actually do the request, and get the response
        try:
            r = requests.post(url=url, data=monitor, headers=headers, params=params, auth=(username, apikey))
            if r.status_code != requests.codes.ok:
                r.raise_for_status()
        except requests.exceptions.HTTPError as err:
            if r.status_code > 200:
                responseText = r.text
                sys.stderr.write("\nDeployment error: " + responseText)
                return { "status": "error", "message": responseText }
            else:
                sys.stderr.write("\nError in HTTP connection")
                return { "status": "error", "message": "Error in HTTP connection." }
        except Exception as err:
            sys.stderr.write("\nDeployment error: " + str(err))
            return { "status": "error", "message": str(err) }
        rsp = r.text
        # clean up after we're done
        f.close()
        os.unlink(f.name)
        reply = {
            "status": "OK",
            "message": "Model successfully uploaded. Your model will begin building momentarily. Please see %s for more details" % self.base_uri
        }
        return reply

class Yhat(API):

    """
    A connection to the Yhat API

    Parameters
    ----------
    username: string
        your Yhat username
    apikey: string
        your Yhat apikey
    uri: string
        your Yhat server URL (i.e. http://cloud.yhathq.com/)
    """

    def __init__(self, username, apikey, uri):
        self.username = username
        self.apikey = apikey
        if uri.endswith("/") is False:
            uri += "/"
        self.ws = None
        self.base_uri = uri
        self.headers = {'Content-Type': 'application/json'}
        self.q = {"username": self.username, "apikey": apikey}
        if self.base_uri != BASE_URI:
            e = self._authenticate()
            if e is not None:
                raise Exception("Failed to authenticate: {}".format(e))

    def _check_obj_size(self, obj):
        """
        Checks if an object is greater than 50 MB

        Parameters
        ----------
        obj: object
            any object

        Returns
        -------
        boolean
        """
        if self.base_uri != BASE_URI:
            # not deploying to the cloud so models can be as big as you want
            if sys.getsizeof(obj) > 52428800:
                return False
        elif sys.getsizeof(obj) > 52428800:
            raise Exception("Sorry, your file is too big for a free account.")

        return True

    def _authenticate(self):
        """
        Returns
        -------
        error: None or Exception
            verifies your API credentials are valid
        """
        try:
            response = self._post('verify', self.q, {})
            error = response["success"]
            return None
        except Exception as e:
            return e

    def _convert_to_json(self, data):
        """
        Tries to convert data into a JSON frieldy dict.

        Parameters
        ----------
        data: object
            data to be converted to valid JSON

        Returns
        -------
        data: object
            same data, but compatible with JSON
        """
        try:
            import pandas as pd
        except ImportError:
            return data

        if isinstance(data, pd.DataFrame):
            data_values = data.transpose().to_json(
                orient='values', date_format='iso')
            data_values = json.loads(data_values)
            try:
                from collections import OrderedDict
                data = OrderedDict(list(zip(data.columns, data_values)))
            except ImportError:
                data = dict(list(zip(data.columns, data_values)))

        data = json.dumps(data)
        return data

    def predict(self, model, data, model_owner=None, raw_input=False):
        """
        Executes a single prediction via the Yhat API.

        Parameters
        ----------
        model: string
            the name of your model
        data: dictionary/data frame
            data required to make a single prediction. this can be a dict or
            a dataframe
        model_owner: string
            username of the model owner for shared models. optional
        raw_input: bool
            whether or not to attempt to parse the incoming JSON. for calling R
            models only

        Returns
        -------
        data: dictionary
            data returned by a model
        """
        data = self._convert_to_json(data)
        q = self.q
        if raw_input==True:
            q['raw_input'] = True
        model_user = self.username if model_owner is None else model_owner
        if self.base_uri != BASE_URI:
            endpoint = "%s/models/%s/" % (model_user, model)
        else:
            data = {"data": data}
            endpoint = 'predict'
        return self._post(endpoint, q, data)

    def _extract_model(self, name, model, session, autodetect, is_tensorflow=False, verbose=0):
        """
        Extracts source code and any objects required to deploy the model.

        Parameters
        ----------
        name: string
            name of your model
        model: YhatModel
            an instance of a Yhat model
        session: globals()
            your Python's session variables (i.e. "globals()")
        verbose: int
            log level
        """
        code = ""
        print("extracting model")

        levels = {
            0: logging.WARNING,
            1: logging.INFO,
            2: logging.DEBUG
        }
        if verbose > 2:
            verbose = 2

        logging.basicConfig(format='[%(levelname)s]: %(message)s', level=levels[verbose])
        bundle = save_function(model, session)
        bundle["largefile"] = True
        bundle["username"] = self.username
        bundle["language"] = "python"
        bundle["modelname"] = name
        bundle["className"] = model.__name__
        bundle["code"] = code + "\n" + bundle.get("code", "")

        # REQUIREMENTS
        if DETECT_REQUIREMENTS and autodetect and getImplicitRequirements:
            requirements = getImplicitRequirements(model, session)
        elif getExplicitRequirements:
            requirements = getExplicitRequirements(model, session)
        else:
            requirements = "\n".join(getattr(model, "REQUIREMENTS", []))

        bundle["reqs"] = requirements

        submodules = detect_explicit_submodules(model)
        # MODULES
        modules = bundle.get("modules", [])
        existing_filenames = [os.path.join(m['parent_dir'], m['name']) for m in modules]
        for submodule in submodules:
            new_filename = os.path.join(submodule['parent_dir'], submodule['name'])
            if new_filename not in existing_filenames:
                modules.append(submodule)

        if modules:
            print("model source files")
            for module in modules:
                name = module["name"]
                parent_dir = module.get("parent_dir", "")
                if parent_dir != "":
                    name = os.path.join(parent_dir, name)
                print(" [+]", name)

        # OBJECTS
        objects = bundle.get("objects", {})
        if objects:
            print("model variables")
            for name, pkl in objects.items():
                if name=='__tensorflow_session':
                    continue
                try:
                    try:
                        obj = terragon.loads_from_base64(pkl)
                    except:
                        try:
                            obj = terragon.loads_spark_from_base64(session['sc'], pkl)
                        except:
                            obj = terragon.load_pom_from_base64(pkl)

                    t = type(obj)
                    del obj
                except Exception as e:
                    print("ERROR pickling object:", name)
                    raise e
                size = 3. * float(len(pkl)) / 4.
                print(" [+]", name, t, sizeof_fmt(size))

        if is_tensorflow==True:
            bundle['objects']['__tensorflow_session'] = terragon.sparkle.save_tensorflow_graph(session['sess'])

        return bundle

    def deploy(self, name, model, session, sure=False, packages=[], patch=None, dry_run=False, verbose=0, autodetect=True, is_tensorflow=False):
        """
        Deploys your model to a Yhat server

        Parameters
        ----------
        name: string
            name of your model
        model: YhatModel
            an instance of a Yhat model
        session: globals()
            your Python's session variables (i.e. "globals()")
        packages: list (deprecated in ScienceOps 2.7.x)
            this is being deprecated in favor of custom runtime images
        sure: boolean
            if true, then this will force a deployment (like -y in apt-get).
            if false or blank, this will ask you if you're sure you want to
            deploy
        verbose: int
            Relative amount of logging info to display (higher = more logs)
        autodetect: flag for using the requirement auto-detection feature.
            if False, you should explicitly state the packages required for
            your model, or it may not run on the server.
        """
        # first let's check and make sure the user actually wants to deploy
        # a new version
        if not re.match("^[A-Za-z0-9_]+$", name):
            raise Exception("Model name must only contain: [A-Za-z0-9_]")
        if not isinstance(packages, list):
            raise Exception(
                "`packages` must be a list of ubuntu packages to install")
        if (not sure) and (not dry_run):
            sure = str(input("Are you sure you want to deploy? (y/N): "))
            if sure.lower() != "y":
                print("Deployment canceled")
                sys.exit()
        bundle = self._extract_model(name, model, session, verbose=verbose, autodetect=autodetect, is_tensorflow=is_tensorflow)
        bundle['packages'] = packages
        if isinstance(patch, string_types) == True:
            patch = "\n".join([line.strip() for line in patch.strip().split('\n')])
            bundle['code'] = patch + "\n" + bundle['code']

        if dry_run:
            return {"status": "ok", "info": "dry run complete"}, bundle
        if self._check_obj_size(bundle) is False:
            # we're not going to deploy; model is too big, but let's give the
            # user the option to upload it manually
            raise Exception("Model is too large to deploy over HTTP")
        else:
            # upload the model to the server
            data = self._post_file("deployer/model/large", self.q, bundle, pb=True)
            return data

    def deploy_tensorflow(self, name, model, session, sess, sure=False, packages=[], patch=None, dry_run=False, verbose=0, autodetect=True):
        """
        Deploys a TensorFlow model to a Yhat server. This is a special case of deploy.

        Parameters
        ----------
        name: string
            name of your model
        model: YhatModel
            an instance of a Yhat model
        session: globals()
            your Python's session variables (i.e. "globals()")
        sess: tensorflow.Session, tensorflow.InteractiveSession
            your tensorflow session. this is typically `sess`
        packages: list (deprecated in ScienceOps 2.7.x)
            this is being deprecated in favor of custom runtime images
        sure: boolean
            if true, then this will force a deployment (like -y in apt-get).
            if false or blank, this will ask you if you're sure you want to
            deploy
        verbose: int
            Relative amount of logging info to display (higher = more logs)
        autodetect: flag for using the requirement auto-detection feature.
            if False, you should explicitly state the packages required for
            your model, or it may not run on the server.
        """

        try:
            model.setup_tf
        except:
            raise Exception("tensorflow models must have a `setup_tf` function")

        if 'sess' not in session:
            session['sess'] = sess

        src = "\n".join(inspect.getsource(model.setup_tf).split('\n')[1:])
        patch = "print('loading tensorflow session...')\n"
        patch += reindent(src)
        patch += "sess, _ = __terragon.sparkle.load_tensorflow_graph(__bundle['objects']['__tensorflow_session'])\n"
        patch += "print('done!')\n\n"

        return self.deploy(name, model, session, sure=sure, packages=packages,
            patch=patch, dry_run=dry_run, verbose=verbose, autodetect=autodetect, is_tensorflow=True)

    def deploy_spark(self, name, model, session, sc, sure=False, packages=[], patch=None, dry_run=False, verbose=0, autodetect=True):
        """
        Deploys a Spark model to a Yhat server. This is a special case of deploy.

        Parameters
        ----------
        name: string
            name of your model
        model: YhatModel
            an instance of a Yhat model
        session: globals()
            your Python's session variables (i.e. "globals()")
        sc: SparkContext
            your SparkContext. this is typically `sc`
        packages: list (deprecated in ScienceOps 2.7.x)
            this is being deprecated in favor of custom runtime images
        sure: boolean
            if true, then this will force a deployment (like -y in apt-get).
            if false or blank, this will ask you if you're sure you want to
            deploy
        verbose: int
            Relative amount of logging info to display (higher = more logs)
        autodetect: flag for using the requirement auto-detection feature.
            if False, you should explicitly state the packages required for
            your model, or it may not run on the server.
        """
        if 'sc' not in session:
            session['sc'] = sc

        patch = "from pyspark import SparkContext\n"
        patch += "sc = SparkContext()\n"
        patch += "sc.setLogLevel('ERROR')\n"

        return self.deploy(name, model, session, sure=sure, packages=packages,
            patch=patch, dry_run=dry_run, verbose=verbose, autodetect=autodetect)

def zlib_compress(data, to):
    step = 4 << 20 # 4MiB
    c = zlib.compressobj()
    data = bytes(data, "utf-8")
    for i in range(0, len(data), step):
        to.write(c.compress(data[i:i+step]))

    to.write(c.flush())



import json

def is_valid_json(data):
    try:
        json.dumps(data)
        return True
    except Exception as e:
        msg = """Whoops. The data you're trying to send could not be
converted into JSON. If the data you're attempting to send includes a numpy
array, try casting it to a list (x.tolist()), or consider structuring your data
as a pandas DataFrame. If you're still having trouble, please contact:
{URL}.""".format(URL="support@yhathq.com")
        print(msg)
        return False


def sizeof_fmt(num, suffix='B'):
    for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)

def create_tests(df, output_file, columns=None):
    """
    Creates ScienceOps compatible (JSON Line Format) smoke test files from your dataframes.
    Once created, these files can be uploaded directly to your model via ScienceOps --> '/models/{model}/unit-tests'

    Parameters
    ----------
    df: dataframe
        the dataframe you wish to pull test inputs from
    output_file: string
        the name you will give your exported test file
    columns: array of strings
        the columns in your df which will be made into inputs

    Example
    -------
    Let's say you've built a book recommender model that takes 2 inputs - title & page count:

    from yhat.utils import create_tests
    books =  pd.read_csv('./book_data.csv')
    create_tests(books, "book_model_inputs.ldjson", columns=["title","pages"])
    * That's it! *
    --> book_model_inputs.ldjson {"title":"harry potter","pages":"800"}\\n{"title":"war and peace","pages":"875"}\\n{"title":"lord of the rings","pages":"500"}

      """
    # if the user doesn't specify columns, then we'll assume they want them all!
    if columns is None:
        columns = df.columns
    with open(output_file, "wb") as f:
        for _, row in df[columns].iterrows():
            f.write(row.to_json() + '\n')



from __future__ import absolute_import
import unittest

from pip._vendor.pkg_resources import Requirement, RequirementParseError
from yhat.requirements import parseUserRequirementsList

class RequirementsTest(unittest.TestCase):


    def testPipFreeze(self):
        """
        testPipFreeze will test that we can parse a requirements.txt file
        from pip freeze
        """
        # Set the requirments to our test requirments list
        REQUIREMENTS = ['./testdata/realtime/test_reqs.txt']
        # Run this through the parser
        userReqs = parseUserRequirementsList(REQUIREMENTS)
        assertReqs = [Requirement.parse('scikit-learn==0.16.1'),
             Requirement.parse('pandas==0.17.1'),
             Requirement.parse('dill==0.2.5'),
             Requirement.parse('terragon==0.3.0'),
             Requirement.parse('progressbar2==3.10.1'),
             Requirement.parse('requests'),
             Requirement.parse('requests-toolbelt')]
        self.assertTrue(userReqs, assertReqs)

    def testCondaList(self):
        """
        testCondaList will test that we can parse a requirements.txt file
        that comes from conda list
        """
        pass

    def testRequirmentsString(self):
        """
        testRequirmentsString will test a normal requirments list
        """
        REQUIREMENTS = ['scikit-learn==0.16.1', 'pandas==0.17.1']
        userReqs = parseUserRequirementsList(REQUIREMENTS)
        assertReqs = [Requirement.parse('scikit-learn==0.16.1'),
             Requirement.parse('pandas==0.17.1')]
        self.assertTrue(userReqs, assertReqs)

if __name__ == '__main__':
    unittest.main()



import json
import logging
import os
import os.path
import re
import sys
import tarfile
from builtins import input

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin
try:
    import StringIO as io
except ImportError:
    import io as io

import requests
from requests_toolbelt.multipart.encoder import MultipartEncoderMonitor, MultipartEncoder
from progressbar import ProgressBar, Percentage, Bar, FileTransferSpeed, ETA

from .deployment.save_session import save_function

class BatchJob(object):

    def __init__(self, name, **kwargs):
        if not re.match("^[a-zA-Z0-9_]+$", name):
            raise ValueError(
                "Job name must contain only [a-zA-Z0-9_]. Got: {}".format(name)
            )
        self.name = name
        for key in ["username", "apikey", "url"]:
            if not key in kwargs:
                raise ValueError("{} not specified".format(key))
            setattr(self, key, kwargs[key])

    def __create_bundle_tar(self, bundle, filename):
        logging.debug("creating batch job tarfile...")
        # Write a bundle file
        f = open('bundle.json', 'w')
        f.write(bundle)
        f.close()

        # make sure old files are gone from previous job
        if os.path.isfile(filename):
            os.remove(filename)

        archive = tarfile.open(filename, "w:gz")
        archive.add('bundle.json')

        if os.path.isfile("yhat.yaml"):
            archive.add("yhat.yaml")
        if os.path.isfile("requirements.txt"):
            archive.add("requirements.txt")
        archive.close()
        # remove the bundle file
        os.unlink('bundle.json')

    def __post_file(self, filename, url, username, job_name, apikey):
        logging.debug("sending batch job to server...")
        def createCallback(encoder):
            # Stuff for progress bar setup
            widgets = ['Transfering Model: ', Bar(), Percentage(), ' ', ETA(), ' ', FileTransferSpeed()]
            pbar = ProgressBar(max_value=encoder.len, widgets=widgets).start()
            def callback(monitor):
                current = monitor.bytes_read
                pbar.update(current)
            return callback

        # Create the Multi-Part encoder
        data = open(filename, "rb")
        encoder = MultipartEncoder(
            fields={'job_name': job_name, 'job': (filename, data, 'application/x-tar')}
        )

        headers = {
            'Content-Type': encoder.content_type,
        }
        callback = createCallback(encoder)
        monitor = MultipartEncoderMonitor(encoder, callback)

        # Actually do the request, and get the response
        try:
            r = requests.post(url=url, data=monitor, headers=headers, auth=(username, apikey))
            if r.status_code != requests.codes.ok:
                r.raise_for_status()
        except requests.exceptions.HTTPError as err:
            text = r.text
            sys.stderr.write("\nServer error: {}".format(text))
            return
        except Exception as e:
            sys.stderr.write("\nError: {}".format(e))
            return
        response_text = r.text

    def deploy(self, session, sure=False, verbose=0):
        """
        Deploy your batch model to the yhat server

        Parameters
        -----
        session: globals()
            your Python's session variables (i.e. "globals()")
        sure: boolean
            if true, then this will force a deployment (like -y in apt-get).
            if false or blank, this will ask you if you're sure you want to
            deploy
        verbose: int
            Relative amount of logging info to display (higher = more logs)
        """
        # Setup logging
        levels = {
            0: logging.WARNING,
            1: logging.INFO,
            2: logging.DEBUG
        }
        if verbose > 2:
            verbose = 2
        logging.basicConfig(format='[%(levelname)s]: %(message)s', level=levels[verbose])

        bundle = save_function(self.__class__, session)
        bundle["class_name"] = self.__class__.__name__
        bundle["language"] = "python"
        bundle_str = json.dumps(bundle)
        filename = ".tmp_yhat_job.tar.gz"
        self.__create_bundle_tar(bundle_str, filename)
        url = urljoin(self.url, "/batch/deploy")
        print(("deploying batch job to: " +  str(url)))
        if not sure:
            sure = input("Are you sure you want to deploy? (y/N): ")
            if sure.lower() != "y":
                print("Deployment canceled")
                sys.exit()
        self.__post_file(filename, url, self.username, self.name, self.apikey)
        os.remove(filename)



import json
import base64
import os
import re
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
from builtins import input


def has():
    """
    Checks if the user has saved their credentials.
    """
    return os.path.isfile(os.path.join(os.environ['HOME'], '.yhat', '.config'))


def setup():
    """
    Prompts the user for their credentials and the saves them to a Yhat "dot"
    file.
    """
    _username = ""
    _apikey = ""
    _server = " [%s]" % "http://cloud.yhathq.com"
    if has():
        creds = read()
        _username = " [%s]" % creds["username"]
        _apikey = " [%s]" % creds["apikey"]
        _server = " [%s]" % creds["server"]
    username = input("Yhat username" + _username + ": ")
    apikey = input("Yhat apikey" + _apikey + ": ")
    server = input("Yhat server" + _server + ": ")

    if username == "":
        username = re.search(r"[^[]*\[([^]]*)\]", _username).group(1)
    if apikey == "":
        apikey = re.search(r"[^[]*\[([^]]*)\]", _apikey).group(1)

    if server == "":
        server = re.search(r"[^[]*\[([^]]*)\]", _server).group(1)
    else:
        if not "http://" in server and not "https://" in server:
            server = "http://" + server
        o = urlparse(server)
        server = "%s://%s" % (o.scheme, o.netloc)

    # create the directory if it doesn't exist
    yhat_dir = os.path.join(os.environ['HOME'], '.yhat')
    if not os.path.exists(yhat_dir):
        os.makedirs(yhat_dir)

    with open(os.path.join(yhat_dir, '.config'), 'w') as f:
        data = json.dumps({"username": username,
                           "apikey": apikey,
                           "server": server})
        data = base64.encodestring(data)
        f.write(data)


def read():
    """
    Extracts credentials from a "dot" file

    Returns
    =======
    credentials: dict
        your credentials in form:
        {"username": "YOUR USERNAME", "apikey": "YOUR APKIKEY}"
    """
    data = open(os.path.join(os.environ['HOME'], '.yhat', '.config')).read()
    return json.loads(base64.decodestring(data))



import unittest
import numpy as np
from yhat import Yhat, YhatModel
from yhat.submodules import detect_explicit_submodules
import json

class TestModel(YhatModel):
    FILES = [
        "sub-sub-modules/run.py"
    ]

class TestYhatJson(unittest.TestCase):

    def test_detect_explicit_submodule(self):
        submodules = detect_explicit_submodules(TestModel)
        has_run_py = False
        for submodule in submodules:
            if submodule['name']=="run.py":
                has_run_py = True
        self.assertTrue(has_run_py)

    def test_detect_submodule_in_deployment(self):
        yh = Yhat("greg", "test", "http://api.yhathq.com/")
        yh.deploy("TestModel", TestModel, globals(), sure=True, dry_run=True)

if __name__=="__main__":
    unittest.main()
